#include "ELF.h"
#include "InitRD.h"
#include "Log.h"
#include "arch/CPU.h"
#include "arch/MMU.h"
#include "arch/PCI.h"
#include "arch/Timer.h"
#include "boot/Init.h"
#include "config.h"
#include "fs/tmpfs/FileSystem.h"
#include "memory/Heap.h"
#include "memory/KernelVM.h"
#include "memory/MemoryManager.h"
#include "thread/Scheduler.h"
#include <luna/CString.h>
#include <luna/Result.h>
#include <luna/Units.h>

void heap_thread()
{
    CPU::disable_interrupts();
    dump_heap_usage();
    kdbgln("Kernel uses %lu vm pages", KernelVM::used() / ARCH_PAGE_SIZE);
    kernel_exit();
}

void reap_thread()
{
    while (true)
    {
        CPU::disable_interrupts();
        auto dying_threads = Scheduler::check_for_dying_threads();
        CPU::enable_interrupts();

        dying_threads.consume([](Thread* thread) { Scheduler::reap_thread(thread); });

        kernel_sleep(250);
    }
}

Result<void> init()
{
    kinfoln("Starting Moon %s, built on %s at %s", MOON_VERSION, __DATE__, __TIME__);

    kinfoln("Current platform: %s", CPU::platform_string());

    kinfoln("Current processor: %s", CPU::identify().value_or("(unknown)"));

    Timer::init();

    kinfoln("Total memory:    %s", to_dynamic_unit(MemoryManager::total()).release_value().chars());
    kinfoln("Free memory:     %s", to_dynamic_unit(MemoryManager::free()).release_value().chars());
    kinfoln("Used memory:     %s", to_dynamic_unit(MemoryManager::used()).release_value().chars());
    kinfoln("Reserved memory: %s", to_dynamic_unit(MemoryManager::reserved()).release_value().chars());

    Thread::init();
    Scheduler::init();

    VFS::root_fs = TRY(TmpFS::FileSystem::create());

    TarStream::Entry entry;
    while (TRY(g_initrd.read_next_entry(entry)))
    {
        if (entry.type == TarStream::EntryType::RegularFile)
        {
            kinfoln("Found file %s in initial ramdisk, of size %s and mode %#ho", entry.name,
                    to_dynamic_unit(entry.size).release_value().chars(), entry.mode);

            if (!strcmp(entry.name, "bin/app")) { TRY(Scheduler::new_userspace_thread(entry, g_initrd)); }
        }
    }

    TRY(Scheduler::new_kernel_thread(heap_thread));
    TRY(Scheduler::new_kernel_thread(reap_thread));

    PCI::scan(
        [](const PCI::Device& device) {
            kinfoln("Found PCI mass storage device %.4x:%.4x, at address %u:%u:%u", device.id.vendor, device.id.device,
                    device.address.bus, device.address.slot, device.address.function);
        },
        { .klass = 1 });

    CPU::platform_finish_init();

    // Disable console logging before transferring control to userspace.
    setup_log(log_debug_enabled(), log_serial_enabled(), false);

    CPU::enable_interrupts();

    return {};
}

[[noreturn]] void init_wrapper()
{
    auto rc = init();
    if (rc.has_error()) kerrorln("Runtime error: %s", rc.error_string());
    CPU::idle_loop();
}

// FIXME: Add a guard page to make sure the stack doesn't end up in random kernel memory. Also reclaim this memory after
// leaving the init task.
extern "C" [[noreturn]] void _start()
{
    Init::check_magic();
    Init::early_init();
    Stack stack { MemoryManager::alloc_for_kernel(8, MMU::ReadWrite | MMU::NoExecute).value(), 8 * ARCH_PAGE_SIZE };
    CPU::bootstrap_switch_stack(stack.top(), (void*)init_wrapper);
}
